from math import pi, ceil, sin, cos, radians
from fontTools.misc.vector import Vector


def buildSweepGradientPatches(
    colorLine,
    center,
    radius,
    startAngle,
    endAngle,
    useGouraudShading,
    maxAngle=None,
):
    """Provides colorful triangular patches that mimic a sweep gradient.

    Together the patches approximate an angular section of a disk of center
    'center' and radius 'radius'.
    The patches respect the color line provided by 'colorLine'.
    The angular section is beetween 'startAngle' and 'endAngle'.

    For use, in particular, in the Cairo and CoreGraphics backends, since these
    libraries lack the sweep gradient feature.

    useGouraudShading -- If True, build a lot of skinny triangles, expected to
    be constant or Gouraud shaded. If False, build fewer degenerate Coons patches
    (outer boundary is rounded), expected to be used in a "mesh gradient"

    Optional keyword arguments:
    maxAngle -- largest desired angular extent of a single triangular patch."""
    patches = []
    # generate a fan of 'triangular' bezier patches, with center 'center' and radius 'radius'
    if maxAngle is None:
        if useGouraudShading:
            maxAngle = pi / 360.0
        else:
            maxAngle = pi / 8.0
    else:
        maxAngle = max(min(maxAngle, pi / 2), pi / 360)
    if useGouraudShading:
        # Use a slightly larger radius to make sure that disk with the original
        # radius completely fits within the straight-edged triangles that we
        # will generate
        radius = radius / cos(maxAngle / 2)
    n = len(colorLine)
    center = Vector(center)
    for i in range(n - 1):
        a0, col0 = colorLine[i + 0]
        a1, col1 = colorLine[i + 1]
        if a0 == a1:
            continue  # two equal stopOffset are used to add color discontinuities. Nothing too draw
        col0 = Vector(col0)
        col1 = Vector(col1)
        a0 = radians(startAngle + a0 * (endAngle - startAngle))
        a1 = radians(startAngle + a1 * (endAngle - startAngle))
        numSplits = int(ceil((a1 - a0) / maxAngle))
        p0 = Vector((cos(a0), sin(a0)))
        color0 = col0
        for a in range(numSplits):
            k = (a + 1.0) / numSplits
            angle1 = a0 + k * (a1 - a0)
            color1 = col0 + k * (col1 - col0)
            p1 = Vector((cos(angle1), sin(angle1)))
            P0 = center[0] + radius * p0[0], center[1] + radius * p0[1]
            P1 = center[0] + radius * p1[0], center[1] + radius * p1[1]
            # draw patch
            if useGouraudShading:
                patches.append(((P0, color0), (P1, color1)))
            else:
                # compute cubic Bezier antennas (control points) so as to approximate the circular arc p0-p1
                A = (p0 + p1).normalized()
                U = Vector((-A[1], A[0]))  # tangent to circle at A
                C0 = A + ((p0 - A).dot(p0) / U.dot(p0)) * U
                C1 = A + ((p1 - A).dot(p1) / U.dot(p1)) * U
                C0 = center + radius * (C0 + 0.33333 * (C0 - p0))
                C1 = center + radius * (C1 + 0.33333 * (C1 - p1))
                patches.append(((P0, color0), C0, C1, (P1, color1)))
            # move to next patch
            p0 = p1
            color0 = color1
    return patches



from contextlib import contextmanager
from math import sqrt
import os
from fontTools.pens.basePen import BasePen
from fontTools.ttLib.tables.otTables import CompositeMode, ExtendMode
from CoreFoundation import CFDataCreateMutable
import Quartz as CG
from .base import Canvas, Surface
from .sweepGradient import buildSweepGradientPatches


_compositeModeMap = {
    CompositeMode.CLEAR: CG.kCGBlendModeClear,
    CompositeMode.SRC: CG.kCGBlendModeCopy,
    CompositeMode.DEST: CG.kCGBlendModeNormal,  # This is wrong, but is worked around in canvas.compositeMode()
    CompositeMode.SRC_OVER: CG.kCGBlendModeNormal,
    CompositeMode.DEST_OVER: CG.kCGBlendModeDestinationOver,
    CompositeMode.SRC_IN: CG.kCGBlendModeSourceIn,
    CompositeMode.DEST_IN: CG.kCGBlendModeDestinationIn,
    CompositeMode.SRC_OUT: CG.kCGBlendModeSourceOut,
    CompositeMode.DEST_OUT: CG.kCGBlendModeDestinationOut,
    CompositeMode.SRC_ATOP: CG.kCGBlendModeSourceAtop,
    CompositeMode.DEST_ATOP: CG.kCGBlendModeDestinationAtop,
    CompositeMode.XOR: CG.kCGBlendModeXOR,
    CompositeMode.PLUS: CG.kCGBlendModePlusLighter,
    CompositeMode.SCREEN: CG.kCGBlendModeScreen,
    CompositeMode.OVERLAY: CG.kCGBlendModeOverlay,
    CompositeMode.DARKEN: CG.kCGBlendModeDarken,
    CompositeMode.LIGHTEN: CG.kCGBlendModeLighten,
    CompositeMode.COLOR_DODGE: CG.kCGBlendModeColorDodge,
    CompositeMode.COLOR_BURN: CG.kCGBlendModeColorBurn,
    CompositeMode.HARD_LIGHT: CG.kCGBlendModeHardLight,
    CompositeMode.SOFT_LIGHT: CG.kCGBlendModeSoftLight,
    CompositeMode.DIFFERENCE: CG.kCGBlendModeDifference,
    CompositeMode.EXCLUSION: CG.kCGBlendModeExclusion,
    CompositeMode.MULTIPLY: CG.kCGBlendModeMultiply,
    CompositeMode.HSL_HUE: CG.kCGBlendModeHue,
    CompositeMode.HSL_SATURATION: CG.kCGBlendModeSaturation,
    CompositeMode.HSL_COLOR: CG.kCGBlendModeColor,
    CompositeMode.HSL_LUMINOSITY: CG.kCGBlendModeLuminosity,
}


class CoreGraphicsPathPen(BasePen):
    def __init__(self):
        super().__init__(None)
        self.path = CG.CGPathCreateMutable()

    def _moveTo(self, pt):
        CG.CGPathMoveToPoint(self.path, None, *pt)

    def _lineTo(self, pt):
        CG.CGPathAddLineToPoint(self.path, None, *pt)

    def _curveToOne(self, pt1, pt2, pt3):
        CG.CGPathAddCurveToPoint(self.path, None, *pt1, *pt2, *pt3)

    def _qCurveToOne(self, pt1, pt2):
        CG.CGPathAddQuadCurveToPoint(self.path, None, *pt1, *pt2)

    def _closePath(self):
        CG.CGPathCloseSubpath(self.path)


class CoreGraphicsCanvas(Canvas):
    def __init__(self, context):
        self.context = context
        self.clipIsEmpty = None

    @staticmethod
    def newPath():
        return CoreGraphicsPathPen()

    @contextmanager
    def savedState(self):
        clipIsEmpty = self.clipIsEmpty
        CG.CGContextSaveGState(self.context)
        yield
        CG.CGContextRestoreGState(self.context)
        self.clipIsEmpty = clipIsEmpty

    @contextmanager
    def compositeMode(self, compositeMode):
        CG.CGContextSaveGState(self.context)
        if compositeMode == CompositeMode.DEST:
            # Workaround for CG not having a blend mode corresponding
            # with CompositeMode.DEST. Setting alpha to 0 should be
            # equivalent.
            CG.CGContextSetAlpha(self.context, 0.0)
        else:
            CG.CGContextSetBlendMode(self.context, _compositeModeMap[compositeMode])
        CG.CGContextBeginTransparencyLayer(self.context, None)
        yield
        CG.CGContextEndTransparencyLayer(self.context)
        CG.CGContextRestoreGState(self.context)

    def transform(self, transform):
        CG.CGContextConcatCTM(self.context, transform)

    def clipPath(self, path):
        if CG.CGPathGetBoundingBox(path.path) == CG.CGRectNull:
            # The path is empty, which causes *no* clip path to be set,
            # which in turn would cause the entire canvas to be filled,
            # so let's prevent that with a flag.
            self.clipIsEmpty = True
        else:
            self.clipIsEmpty = False
            CG.CGContextAddPath(self.context, path.path)
            CG.CGContextClip(self.context)

    def drawPathSolid(self, path, color):
        if self.clipIsEmpty:
            return
        CG.CGContextAddPath(self.context, path.path)
        CG.CGContextSetRGBFillColor(self.context, *color)
        CG.CGContextFillPath(self.context)

    def drawPathLinearGradient(
        self, path, colorLine, pt1, pt2, extendMode, gradientTransform
    ):
        if self.clipIsEmpty or CG.CGPathGetBoundingBox(path.path) == CG.CGRectNull:
            return
        colors, stops = _unpackColorLine(colorLine)
        gradient = CG.CGGradientCreateWithColors(None, colors, stops)
        with self.savedState():
            CG.CGContextAddPath(self.context, path.path)
            CG.CGContextClip(self.context)
            self.transform(gradientTransform)
            CG.CGContextDrawLinearGradient(
                self.context,
                gradient,
                pt1,
                pt2,
                CG.kCGGradientDrawsBeforeStartLocation
                | CG.kCGGradientDrawsAfterEndLocation,
            )

    def drawPathRadialGradient(
        self,
        path,
        colorLine,
        startCenter,
        startRadius,
        endCenter,
        endRadius,
        extendMode,
        gradientTransform,
    ):
        if self.clipIsEmpty or CG.CGPathGetBoundingBox(path.path) == CG.CGRectNull:
            return
        colors, stops = _unpackColorLine(colorLine)
        gradient = CG.CGGradientCreateWithColors(None, colors, stops)
        with self.savedState():
            CG.CGContextAddPath(self.context, path.path)
            CG.CGContextClip(self.context)
            self.transform(gradientTransform)
            CG.CGContextDrawRadialGradient(
                self.context,
                gradient,
                startCenter,
                startRadius,
                endCenter,
                endRadius,
                CG.kCGGradientDrawsBeforeStartLocation
                | CG.kCGGradientDrawsAfterEndLocation,
            )

    def drawPathSweepGradient(
        self,
        path,
        colorLine,
        center,
        startAngle,
        endAngle,
        extendMode,
        gradientTransform,
    ):
        if self.clipIsEmpty or CG.CGPathGetBoundingBox(path.path) == CG.CGRectNull:
            return
        with self.savedState():
            CG.CGContextAddPath(self.context, path.path)
            CG.CGContextClip(self.context)
            self.transform(gradientTransform)
            # find current path' extent
            (x1, y1), (w, h) = CG.CGContextGetClipBoundingBox(self.context)
            x2 = x1 + w
            y2 = y1 + h
            maxX = max(d * d for d in (x1 - center[0], x2 - center[0]))
            maxY = max(d * d for d in (y1 - center[1], y2 - center[1]))
            R = sqrt(maxX + maxY)
            # compute the triangle fan approximating the sweep gradient
            patches = buildSweepGradientPatches(
                colorLine, center, R, startAngle, endAngle, useGouraudShading=True
            )
            CG.CGContextBeginTransparencyLayer(self.context, None)
            CG.CGContextSetAllowsAntialiasing(self.context, False)
            for (P0, color0), (P1, color1) in patches:
                color = 0.5 * (color0 + color1)
                CG.CGContextMoveToPoint(self.context, center[0], center[1])
                CG.CGContextAddLineToPoint(self.context, P0[0], P0[1])
                CG.CGContextAddLineToPoint(self.context, P1[0], P1[1])
                CG.CGContextSetRGBFillColor(self.context, *color)
                CG.CGContextFillPath(self.context)
            CG.CGContextSetAllowsAntialiasing(self.context, True)
            CG.CGContextEndTransparencyLayer(self.context)

    # TODO: blendMode for PaintComposite


def _unpackColorLine(colorLine):
    colors = []
    stops = []
    for stop, color in colorLine:
        colors.append(CG.CGColorCreateGenericRGB(*color))
        stops.append(stop)
    return colors, stops


class CoreGraphicsPixelSurface(Surface):
    fileExtension = ".png"

    def __init__(self):
        self.context = None

    @contextmanager
    def canvas(self, boundingBox):
        x, y, xMax, yMax = boundingBox
        width = xMax - x
        height = yMax - y
        self._setupCGContext(x, y, width, height)
        yield CoreGraphicsCanvas(self.context)

    def _setupCGContext(self, x, y, width, height):
        rgbColorSpace = CG.CGColorSpaceCreateDeviceRGB()
        self.context = CG.CGBitmapContextCreate(
            None, width, height, 8, 0, rgbColorSpace, CG.kCGImageAlphaPremultipliedFirst
        )
        CG.CGContextTranslateCTM(self.context, -x, -y)

    def saveImage(self, path):
        image = CG.CGBitmapContextCreateImage(self.context)
        saveImageAsPNG(image, path)


class CoreGraphicsPDFSurface(CoreGraphicsPixelSurface):
    fileExtension = ".pdf"

    @contextmanager
    def canvas(self, boundingBox):
        with super().canvas(boundingBox) as canvas:
            CG.CGContextBeginPage(self.context, self._mediaBox)
            yield canvas
            CG.CGContextEndPage(self.context)

    def _setupCGContext(self, x, y, width, height):
        if self.context is None:
            self._mediaBox = ((x, y), (width, height))
            self._data = CFDataCreateMutable(None, 0)
            consumer = CG.CGDataConsumerCreateWithCFData(self._data)
            self.context = CG.CGPDFContextCreate(consumer, self._mediaBox, None)
        return self.context

    def saveImage(self, path):
        CG.CGPDFContextClose(self.context)
        with open(path, "wb") as f:
            f.write(self._data)


def saveImageAsPNG(image, path):
    path = os.path.abspath(path).encode("utf-8")
    url = CG.CFURLCreateFromFileSystemRepresentation(None, path, len(path), False)
    assert url is not None
    dest = CG.CGImageDestinationCreateWithURL(url, "public.png", 1, None)
    assert dest is not None
    CG.CGImageDestinationAddImage(dest, image, None)
    CG.CGImageDestinationFinalize(dest)



from contextlib import contextmanager
import os
from fontTools.pens.basePen import BasePen
from fontTools.ttLib.tables.otTables import CompositeMode, ExtendMode
import skia
from .base import Canvas, Surface


_compositeModeMap = {
    CompositeMode.CLEAR: skia.BlendMode.kClear,
    CompositeMode.SRC: skia.BlendMode.kSrc,
    CompositeMode.DEST: skia.BlendMode.kDst,
    CompositeMode.SRC_OVER: skia.BlendMode.kSrcOver,
    CompositeMode.DEST_OVER: skia.BlendMode.kDstOver,
    CompositeMode.SRC_IN: skia.BlendMode.kSrcIn,
    CompositeMode.DEST_IN: skia.BlendMode.kDstIn,
    CompositeMode.SRC_OUT: skia.BlendMode.kSrcOut,
    CompositeMode.DEST_OUT: skia.BlendMode.kDstOut,
    CompositeMode.SRC_ATOP: skia.BlendMode.kSrcATop,
    CompositeMode.DEST_ATOP: skia.BlendMode.kDstATop,
    CompositeMode.XOR: skia.BlendMode.kXor,
    CompositeMode.PLUS: skia.BlendMode.kPlus,
    CompositeMode.SCREEN: skia.BlendMode.kScreen,
    CompositeMode.OVERLAY: skia.BlendMode.kOverlay,
    CompositeMode.DARKEN: skia.BlendMode.kDarken,
    CompositeMode.LIGHTEN: skia.BlendMode.kLighten,
    CompositeMode.COLOR_DODGE: skia.BlendMode.kColorDodge,
    CompositeMode.COLOR_BURN: skia.BlendMode.kColorBurn,
    CompositeMode.HARD_LIGHT: skia.BlendMode.kHardLight,
    CompositeMode.SOFT_LIGHT: skia.BlendMode.kSoftLight,
    CompositeMode.DIFFERENCE: skia.BlendMode.kDifference,
    CompositeMode.EXCLUSION: skia.BlendMode.kExclusion,
    CompositeMode.MULTIPLY: skia.BlendMode.kMultiply,
    CompositeMode.HSL_HUE: skia.BlendMode.kHue,
    CompositeMode.HSL_SATURATION: skia.BlendMode.kSaturation,
    CompositeMode.HSL_COLOR: skia.BlendMode.kColor,
    CompositeMode.HSL_LUMINOSITY: skia.BlendMode.kLuminosity,
}


_extendModeMap = {
    ExtendMode.PAD: skia.TileMode.kClamp,
    ExtendMode.REPEAT: skia.TileMode.kRepeat,
    ExtendMode.REFLECT: skia.TileMode.kMirror,
}


class SkiaPath(BasePen):
    def __init__(self):
        super().__init__(None)
        self.path = skia.Path()

    def _moveTo(self, pt):
        self.path.moveTo(*pt)

    def _lineTo(self, pt):
        self.path.lineTo(*pt)

    def _curveToOne(self, pt1, pt2, pt3):
        self.path.cubicTo(*pt1, *pt2, *pt3)

    def _qCurveToOne(self, pt1, pt2):
        self.path.quadTo(*pt1, *pt2)

    def _closePath(self):
        self.path.close()


class SkiaCanvas(Canvas):
    def __init__(self, canvas):
        self.canvas = canvas

    @staticmethod
    def newPath():
        return SkiaPath()

    @contextmanager
    def savedState(self):
        self.canvas.save()
        yield
        self.canvas.restore()

    @contextmanager
    def compositeMode(self, compositeMode):
        paint = skia.Paint(BlendMode=_compositeModeMap[compositeMode])
        self.canvas.saveLayer(paint=paint)
        yield
        self.canvas.restore()

    def transform(self, transform):
        matrix = skia.Matrix()
        matrix.setAffine(transform)
        self.canvas.concat(matrix)

    def clipPath(self, path):
        self.canvas.clipPath(path.path, doAntiAlias=True)

    def drawPathSolid(self, path, color):
        paint = skia.Paint(
            AntiAlias=True,
            Color=skia.Color4f(tuple(color)),
            Style=skia.Paint.kFill_Style,
        )
        self.canvas.drawPath(path.path, paint)

    def drawPathLinearGradient(
        self, path, colorLine, pt1, pt2, extendMode, gradientTransform
    ):
        matrix = skia.Matrix()
        matrix.setAffine(gradientTransform)
        colors, stops = _unpackColorLine(colorLine)
        shader = skia.GradientShader.MakeLinear(
            points=[pt1, pt2],
            colors=colors,
            positions=stops,
            mode=_extendModeMap[extendMode],
            localMatrix=matrix,
        )
        self.canvas.drawPath(path.path, skia.Paint(AntiAlias=True, Shader=shader))

    def drawPathRadialGradient(
        self,
        path,
        colorLine,
        startCenter,
        startRadius,
        endCenter,
        endRadius,
        extendMode,
        gradientTransform,
    ):
        matrix = skia.Matrix()
        matrix.setAffine(gradientTransform)
        colors, stops = _unpackColorLine(colorLine)
        shader = skia.GradientShader.MakeTwoPointConical(
            start=startCenter,
            startRadius=startRadius,
            end=endCenter,
            endRadius=endRadius,
            colors=colors,
            positions=stops,
            mode=_extendModeMap[extendMode],
            localMatrix=matrix,
        )
        self.canvas.drawPath(path.path, skia.Paint(AntiAlias=True, Shader=shader))

    def drawPathSweepGradient(
        self,
        path,
        colorLine,
        center,
        startAngle,
        endAngle,
        extendMode,
        gradientTransform,
    ):
        # The following is needed to please the Skia shader, but it's a but fuzzy
        # to me how this affects the spec. Translated from:
        # https://source.chromium.org/chromium/chromium/src/+/master:third_party/skia/src/ports/SkFontHost_FreeType_common.cpp;l=673-686
        startAngle %= 360
        endAngle %= 360
        if startAngle >= endAngle:
            endAngle += 360
        matrix = skia.Matrix()
        matrix.setAffine(gradientTransform)
        colors, stops = _unpackColorLine(colorLine)
        shader = skia.GradientShader.MakeSweep(
            cx=center[0],
            cy=center[1],
            colors=colors,
            positions=stops,
            mode=_extendModeMap[extendMode],
            startAngle=startAngle,
            endAngle=endAngle,
            localMatrix=matrix,
        )
        self.canvas.drawPath(path.path, skia.Paint(AntiAlias=True, Shader=shader))

    # TODO: blendMode for PaintComposite


def _unpackColorLine(colorLine):
    colors = []
    stops = []
    for stop, color in colorLine:
        colors.append(int(skia.Color4f(tuple(color))))
        stops.append(stop)
    return colors, stops


class _SkiaBaseSurface(Surface):
    @contextmanager
    def canvas(self, boundingBox):
        x, y, xMax, yMax = boundingBox
        width = xMax - x
        height = yMax - y
        skCanvas, surfaceData = self._setupSkCanvas(x, y, width, height)
        skCanvas.translate(-x, height + y)
        skCanvas.scale(1, -1)
        yield SkiaCanvas(skCanvas)
        self._finalizeCanvas(surfaceData)


class SkiaPixelSurface(_SkiaBaseSurface):
    fileExtension = ".png"

    def __init__(self):
        self._image = None

    def _setupSkCanvas(self, x, y, width, height):
        surface = skia.Surface(width, height)
        return surface.getCanvas(), surface

    def _finalizeCanvas(self, surface):
        self._image = surface.makeImageSnapshot()

    def saveImage(self, path, format=skia.kPNG):
        self._image.save(os.fspath(path), format)


class SkiaPDFSurface(_SkiaBaseSurface):
    fileExtension = ".pdf"

    def __init__(self):
        self._pictures = []

    def _setupSkCanvas(self, x, y, width, height):
        recorder = skia.PictureRecorder()
        return recorder.beginRecording(width, height), recorder

    def _finalizeCanvas(self, recorder):
        self._pictures.append(recorder.finishRecordingAsPicture())

    def saveImage(self, path):
        stream = skia.FILEWStream(os.fspath(path))
        with skia.PDF.MakeDocument(stream) as document:
            for picture in self._pictures:
                x, y, width, height = picture.cullRect()
                assert x == 0 and y == 0
                with document.page(width, height) as canvas:
                    canvas.drawPicture(picture)
        stream.flush()


class SkiaSVGSurface(SkiaPDFSurface):
    fileExtension = ".svg"

    def saveImage(self, path):
        stream = skia.FILEWStream(os.fspath(path))
        picture = self._pictures[-1]
        canvas = skia.SVGCanvas.Make(picture.cullRect(), stream)
        canvas.drawPicture(picture)
        del canvas  # hand holding skia-python with GC: it needs to go before stream
        stream.flush()



from collections import defaultdict
import importlib


_surfaces = {
    None: {
        "cairo": "blackrenderer.backends.cairo.CairoPixelSurface",
        "coregraphics": "blackrenderer.backends.coregraphics.CoreGraphicsPixelSurface",
        "skia": "blackrenderer.backends.skia.SkiaPixelSurface",
        "svg": "blackrenderer.backends.svg.SVGSurface",
    },
    ".png": {
        "cairo": "blackrenderer.backends.cairo.CairoPixelSurface",
        "coregraphics": "blackrenderer.backends.coregraphics.CoreGraphicsPixelSurface",
        "skia": "blackrenderer.backends.skia.SkiaPixelSurface",
    },
    ".pdf": {
        "cairo": "blackrenderer.backends.cairo.CairoPDFSurface",
        "coregraphics": "blackrenderer.backends.coregraphics.CoreGraphicsPDFSurface",
        "skia": "blackrenderer.backends.skia.SkiaPDFSurface",
    },
    ".svg": {
        "cairo": "blackrenderer.backends.cairo.CairoSVGSurface",
        "skia": "blackrenderer.backends.skia.SkiaSVGSurface",
        "svg": "blackrenderer.backends.svg.SVGSurface",
    },
}


def getSurfaceClass(backendName, imageExtension=None):
    fqName = _surfaces[imageExtension][backendName]
    moduleName, className = fqName.rsplit(".", 1)
    try:
        module = importlib.import_module(moduleName)
    except ModuleNotFoundError:
        return None
    return getattr(module, className)


def listBackends():
    backends = defaultdict(list)
    for suffix in _surfaces:
        if suffix is None:
            continue
        for backendName in _surfaces[suffix]:
            backends[backendName].append(suffix)
    return [
        (backendName, sorted(suffixes)) for backendName, suffixes in backends.items()
    ]



import argparse
import os
import pathlib
import re
from .render import renderText
from .backends import listBackends


backendsAndSuffixes = listBackends()
backendNames = [backendName for backendName, _ in backendsAndSuffixes]

description = f"""\
Render a text string to an image file. Available backends:
"""
for backendName, suffixes in backendsAndSuffixes:
    description += f" {backendName} ({', '.join(suffixes)})"


def main():
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("font", metavar="FONT", type=existingFilePath, help="a font")
    parser.add_argument("text", metavar="TEXT", help="a string")
    parser.add_argument(
        "output",
        metavar="OUTPUT",
        type=outputFilePath,
        help="an output file name, with .png, .pdf or .svg extension, "
        "or '-', to print SVG to stdout",
    )
    parser.add_argument("--font-size", type=float, default=250)
    parser.add_argument("--features", type=parseFeatures)
    parser.add_argument("--variations", type=parseVariations)
    parser.add_argument("--margin", type=float, default=20)
    parser.add_argument(
        "--backend",
        default=None,
        choices=backendNames,
        help="The backend to use -- defaults to skia, except when rendering to "
        ".svg, in which case the svg backend will be used.",
    )
    args = parser.parse_args()
    renderText(
        args.font,
        args.text,
        args.output,
        fontSize=args.font_size,
        margin=args.margin,
        features=args.features,
        variations=args.variations,
        backendName=args.backend,
    )


def existingFilePath(path):
    if not os.path.exists(path):
        raise argparse.ArgumentTypeError(f"file does not exist: '{path}'")
    elif not os.path.isfile(path):
        raise argparse.ArgumentTypeError(f"path is not a file: '{path}'")
    return pathlib.Path(path).resolve()


def outputFilePath(path):
    if path == "-":
        return None
    path = pathlib.Path(path).resolve()
    if path.suffix not in {".png", ".pdf", ".svg"}:
        raise argparse.ArgumentTypeError(
            f"path does not have the right extension; should be .png or .svg: "
            f"'{path}'"
        )
    return path


def parseVariations(string):
    location = {}
    for item in string.split(","):
        axisTag, axisValue = item.split("=")
        axisTag = axisTag.strip()
        axisValue = float(axisValue)
        if len(axisTag) < 0:
            axisTag += " " * (4 - len(axisTag))
        location[axisTag] = axisValue
    return location


feaPat = re.compile(r"(\+|-)?(\w+)(=(\d+))?$")  # kern,-calt,+liga,aalt=2


def parseFeatures(src):
    features = {}
    for part in src.split(","):
        m = feaPat.match(part.strip())
        if m is None:
            raise ValueError(part)
        sign, featureTag, _, altIndex = m.groups()
        if sign == "-":
            value = False
        elif altIndex:
            value = int(altIndex)  # catch ValueError
        else:
            value = True
        features[featureTag] = value
    return features


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from contextlib import contextmanager


class Canvas(ABC):
    @abstractmethod
    def newPath(self):
        ...

    @abstractmethod
    @contextmanager
    def savedState(self):
        ...

    @abstractmethod
    @contextmanager
    def compositeMode(self, compositeMode):
        ...

    @abstractmethod
    def transform(self, transform):
        ...

    @abstractmethod
    def clipPath(self, path):
        ...

    @abstractmethod
    def drawPathSolid(self, path, color):
        ...

    @abstractmethod
    def drawPathLinearGradient(
        self, path, colorLine, pt1, pt2, extendMode, gradientTransform
    ):
        ...

    @abstractmethod
    def drawPathRadialGradient(
        self,
        path,
        colorLine,
        startCenter,
        startRadius,
        endCenter,
        endRadius,
        extendMode,
        gradientTransform,
    ):
        ...

    @abstractmethod
    def drawPathSweepGradient(
        self,
        path,
        colorLine,
        center,
        startAngle,
        endAngle,
        extendMode,
        gradientTransform,
    ):
        ...

    # Generic convenience methods

    def translate(self, x, y):
        self.transform((1, 0, 0, 1, x, y))

    def scale(self, sx, sy=None):
        if sy is None:
            sy = sx
        self.transform((sx, 0, 0, sy, 0, 0))

    def drawRectSolid(self, rect, color):
        self.drawPathSolid(self._rectPath(rect), color)

    def drawRectLinearGradient(self, rect, *args, **kwargs):
        self.drawPathLinearGradient(self._rectPath(rect), *args, **kwargs)

    def drawRectRadialGradient(self, rect, *args, **kwargs):
        self.drawPathRadialGradient(self._rectPath(rect), *args, **kwargs)

    def drawRectSweepGradient(self, rect, *args, **kwargs):
        self.drawPathSweepGradient(self._rectPath(rect), *args, **kwargs)

    def _rectPath(self, rect):
        x, y, w, h = rect
        path = self.newPath()
        path.moveTo((x, y))
        path.lineTo((x, y + h))
        path.lineTo((x + w, y + h))
        path.lineTo((x + w, y))
        path.closePath()
        return path


class Surface(ABC):
    fileExtension = ".png"

    @abstractmethod
    def __init__(self):
        ...

    @abstractmethod
    @contextmanager
    def canvas(self, boundingBox):
        # boundingBox = (xMin, yMin, xMax, yMax)
        ...

    @abstractmethod
    def saveImage(self, path):
        ...



import pathlib
from blackrenderer.font import BlackRendererFont
from blackrenderer.dumpCOLRv1Glyph import dumpCOLRv1Glyph


testDir = pathlib.Path(__file__).resolve().parent
testFont1 = testDir / "data" / "noto-glyf_colr_1.ttf"


expected_output = """\
uni2693
  # PaintColrLayers
  Layers
    - # PaintGlyph
      Format 10
      Paint
          # PaintSolid
          Format 2
          PaletteIndex 2148
          Alpha 1.0
      Glyph glyph39143
    - # PaintGlyph
      Format 10
      Paint
          # PaintSolid
          Format 2
          PaletteIndex 871
          Alpha 1.0
      Glyph glyph39144
    - # PaintGlyph
      Format 10
      Paint
          # PaintSolid
          Format 2
          PaletteIndex 2491
          Alpha 1.0
      Glyph glyph39145
    - # PaintGlyph
      Format 10
      Paint
          # PaintSolid
          Format 2
          PaletteIndex 871
          Alpha 0.89
      Glyph glyph39146
    - # PaintGlyph
      Format 10
      Paint
          # PaintSolid
          Format 2
          PaletteIndex 2491
          Alpha 1.0
      Glyph glyph39147
    - # PaintGlyph
      Format 10
      Paint
          # PaintSolid
          Format 2
          PaletteIndex 871
          Alpha 1.0
      Glyph glyph39148
"""


def test_dump(capsys):
    font = BlackRendererFont(testFont1)
    dumpCOLRv1Glyph(font, "uni2693")
    captured = capsys.readouterr()
    assert expected_output == captured.out



from contextlib import contextmanager
from fontTools.misc.arrayTools import calcBounds
from fontTools.misc.transform import Identity
from fontTools.pens.recordingPen import RecordingPen
from fontTools.pens.transformPen import TransformPen
from .base import Canvas


class PathCollectorCanvas(Canvas):
    def __init__(self):
        self.init()

    def init(self):
        self.paths = []
        self.currentTransform = Identity

    def _addPath(self, path):
        if self.currentTransform != Identity:
            path = transformPath(path, self.currentTransform)
        self.paths.append(path)

    def newPath(self):
        return RecordingPen()

    @contextmanager
    def savedState(self):
        savedTransform = self.currentTransform
        yield
        self.currentTransform = savedTransform

    @contextmanager
    def compositeMode(self, compositeMode):
        yield

    def transform(self, transform):
        self.currentTransform = self.currentTransform.transform(transform)

    def clipPath(self, path):
        self._addPath(path)

    def drawPathSolid(self, path, color):
        self._addPath(path)

    def drawPathLinearGradient(
        self, path, colorLine, pt1, pt2, extendMode, gradientTransform
    ):
        self._addPath(path)

    def drawPathRadialGradient(
        self,
        path,
        colorLine,
        startCenter,
        startRadius,
        endCenter,
        endRadius,
        extendMode,
        gradientTransform,
    ):
        self._addPath(path)

    def drawPathSweepGradient(
        self,
        path,
        colorLine,
        center,
        startAngle,
        endAngle,
        extendMode,
        gradientTransform,
    ):
        self._addPath(path)


class PointCollector:
    def __init__(self):
        self.points = []

    def moveTo(self, pt):
        self.points.append(pt)

    def lineTo(self, pt):
        self.points.append(pt)

    def curveTo(self, *pts):
        self.points.extend(pts)

    qCurveTo = curveTo

    def closePath(self):
        pass

    def endPath(self):
        pass


class BoundsCanvas(PathCollectorCanvas):
    def init(self):
        self.points = []
        self.currentTransform = Identity

    @property
    def bounds(self):
        return calcBounds(self.points)

    def _addPath(self, path):
        points = path.points
        if self.currentTransform != Identity:
            points = self.currentTransform.transformPoints(points)
        self.points.extend(points)

    def newPath(self):
        return PointCollector()


def transformPath(path, transform):
    transformedPath = RecordingPen()
    tpen = TransformPen(transformedPath, transform)
    path.replay(tpen)
    return transformedPath



import pathlib
import pytest
from fontTools.misc.arrayTools import scaleRect, intRect
from blackrenderer.font import BlackRendererFont
from blackrenderer.backends import getSurfaceClass
from blackrenderer.backends.pathCollector import BoundsCanvas, PathCollectorCanvas
from compareImages import compareImages


testDir = pathlib.Path(__file__).resolve().parent
dataDir = testDir / "data"
expectedOutputDir = testDir / "expectedOutput"
tmpOutputDir = testDir / "tmpOutput"
if not tmpOutputDir.exists():
    tmpOutputDir.mkdir()


backends = [
    (name, getSurfaceClass(name)) for name in ["cairo", "coregraphics", "skia", "svg"]
]
backends = [(name, surface) for name, surface in backends if surface is not None]


testFonts = {
    "noto": dataDir / "noto-glyf_colr_1.ttf",
    "mutator": dataDir / "MutatorSans.ttf",
    "twemoji": dataDir / "TwemojiMozilla.subset.default.3299.ttf",
    "more_samples": dataDir / "more_samples-glyf_colr_1.ttf",
}


test_glyphs = [
    ("noto", "uni2693", None),
    ("noto", "uni2694", None),
    ("noto", "u1F30A", None),
    ("noto", "u1F943", None),
    ("mutator", "B", None),
    ("mutator", "D", {"wdth": 1000}),
    ("twemoji", "uni3299", None),
    ("more_samples", "cross_glyph", None),
    ("more_samples", "skew_0_15_center_500.0_500.0", None),
    ("more_samples", "skew_-10_20_center_500.0_500.0", None),
    ("more_samples", "skew_-10_20_center_1000_1000", None),
    ("more_samples", "transform_matrix_1_0_0_1_125_125", None),
    ("more_samples", "transform_matrix_1.5_0_0_1.5_0_0", None),
    ("more_samples", "transform_matrix_0.9659_0.2588_-0.2588_0.9659_0_0", None),
    ("more_samples", "transform_matrix_1.0_0.0_0.6_1.0_-300.0_0.0", None),
    ("more_samples", "clip_box_top_left", None),
    ("more_samples", "clip_box_bottom_left", None),
    ("more_samples", "clip_box_bottom_right", None),
    ("more_samples", "clip_box_top_right", None),
    ("more_samples", "clip_box_center", None),
    ("more_samples", "composite_DEST_OVER", None),
    ("more_samples", "composite_XOR", None),
    ("more_samples", "composite_OVERLAY", None),
    ("more_samples", "composite_SRC_IN", None),
    ("more_samples", "composite_PLUS", None),
    ("more_samples", "composite_LIGHTEN", None),
    ("more_samples", "composite_MULTIPLY", None),
    ("more_samples", "clip_shade_center", None),
    ("more_samples", "clip_shade_top_left", None),
    ("more_samples", "clip_shade_bottom_left", None),
    ("more_samples", "clip_shade_bottom_right", None),
    ("more_samples", "clip_shade_top_right", None),
    ("more_samples", "inset_clipped_radial_reflect", None),
    ("more_samples", "sweep", None),
    ("more_samples", "transformed_sweep", None),
    ("more_samples", "composite_colr_glyph", None),
    ("more_samples", "linear_repeat_0_1", None),
    ("more_samples", "linear_repeat_0.2_0.8", None),
    ("more_samples", "linear_repeat_0_1.5", None),
    ("more_samples", "linear_repeat_0.5_1.5", None),
    ("more_samples", "scale_0.5_1.5_center_500.0_500.0", None),
    ("more_samples", "scale_1.5_1.5_center_500.0_500.0", None),
    ("more_samples", "scale_0.5_1.5_center_0_0", None),
    ("more_samples", "scale_1.5_1.5_center_0_0", None),
    ("more_samples", "scale_0.5_1.5_center_1000_1000", None),
    ("more_samples", "scale_1.5_1.5_center_1000_1000", None),
    ("more_samples", "linear_gradient_extend_mode_pad", None),
    ("more_samples", "linear_gradient_extend_mode_repeat", None),
    ("more_samples", "linear_gradient_extend_mode_reflect", None),
    ("more_samples", "radial_gradient_extend_mode_pad", None),
    ("more_samples", "radial_gradient_extend_mode_repeat", None),
    ("more_samples", "radial_gradient_extend_mode_reflect", None),
    ("more_samples", "rotate_10_center_0_0", None),
    ("more_samples", "rotate_-10_center_1000_1000", None),
    ("more_samples", "rotate_25_center_500.0_500.0", None),
    ("more_samples", "rotate_-15_center_500.0_500.0", None),
    ("more_samples", "skew_25_0_center_0_0", None),
    ("more_samples", "skew_25_0_center_500.0_500.0", None),
    ("more_samples", "skew_0_15_center_0_0", None),
    ("more_samples", "upem_box_glyph", None),
]


@pytest.mark.parametrize("fontName, glyphName, location", test_glyphs)
@pytest.mark.parametrize("backendName, surfaceClass", backends)
def test_renderGlyph(backendName, surfaceClass, fontName, glyphName, location):
    font = BlackRendererFont(testFonts[fontName])
    font.setLocation(location)

    scaleFactor = 1 / 4
    boundingBox = font.getGlyphBounds(glyphName)
    boundingBox = scaleRect(boundingBox, scaleFactor, scaleFactor)
    boundingBox = intRect(boundingBox)

    surface = surfaceClass()
    ext = surface.fileExtension
    with surface.canvas(boundingBox) as canvas:
        canvas.scale(scaleFactor)
        font.drawGlyph(glyphName, canvas)

    fileName = f"glyph_{fontName}_{glyphName}_{backendName}{ext}"
    expectedPath = expectedOutputDir / fileName
    outputPath = tmpOutputDir / fileName
    surface.saveImage(outputPath)
    diff = compareImages(expectedPath, outputPath)
    assert diff < 0.00012, diff


def test_pathCollector():
    font = BlackRendererFont(testFonts["noto"])
    canvas = PathCollectorCanvas()
    font.drawGlyph("uni2693", canvas)
    assert len(canvas.paths) == 6


def test_boundsCanvas():
    font = BlackRendererFont(testFonts["mutator"])
    canvas = BoundsCanvas()
    font.drawGlyph("A", canvas)
    assert (20, 0, 376, 700) == canvas.bounds

    font.setLocation({"wdth": 1000})
    canvas = BoundsCanvas()
    font.drawGlyph("A", canvas)
    assert (50, 0, 1140, 700) == canvas.bounds

    font = BlackRendererFont(testFonts["more_samples"])
    canvas = BoundsCanvas()
    font.drawGlyph("transformed_sweep", canvas)
    assert (317, 154, 1183, 846) == tuple(round(v) for v in canvas.bounds)


vectorBackends = [
    ("cairo", ".pdf"),
    ("cairo", ".svg"),
    ("skia", ".pdf"),
    ("skia", ".svg"),
    ("coregraphics", ".pdf"),
]


@pytest.mark.parametrize("backendName, imageSuffix", vectorBackends)
def test_vectorBackends(backendName, imageSuffix):
    fontName = "noto"
    glyphName = "u1F943"
    surfaceClass = getSurfaceClass(backendName, imageSuffix)
    if surfaceClass is None:
        pytest.skip(f"{backendName} not available")
    assert surfaceClass.fileExtension == imageSuffix

    font = BlackRendererFont(testFonts[fontName])
    boundingBox = font.getGlyphBounds(glyphName)

    surface = surfaceClass()
    with surface.canvas(boundingBox) as canvas:
        font.drawGlyph(glyphName, canvas)
    fileName = f"vector_{fontName}_{glyphName}_{backendName}{imageSuffix}"
    expectedPath = expectedOutputDir / fileName
    outputPath = tmpOutputDir / fileName
    surface.saveImage(outputPath)
    # For now, just be happy the code works.
    # - Cairo PDFs contain the creation date
    # - CoreGraphics PDFs are weirdly different while looking the same
    # assert expectedPath.read_bytes() == outputPath.read_bytes()
    diff = compareImages(expectedPath, outputPath)
    assert diff < 0.00012, diff



from contextlib import contextmanager
import logging
from typing import NamedTuple
from fontTools.misc.transform import Transform
from fontTools.pens.basePen import BasePen
from fontTools.misc import etree as ET
from fontTools.ttLib.tables.otTables import ExtendMode
from .base import Canvas, Surface


logger = logging.getLogger(__name__)


_extendModeMap = {
    ExtendMode.PAD: "pad",
    ExtendMode.REPEAT: "repeat",
    ExtendMode.REFLECT: "reflect",
}


class SVGPath(BasePen):
    def __init__(self, glyphSet=None):
        super().__init__(glyphSet)
        self.segments = []

    def _moveTo(self, pt):
        self.segments.append("M" + formatCoord(pt))

    def _lineTo(self, pt):
        cx, cy = self._getCurrentPoint()
        dx = pt[0] - cx
        dy = pt[1] - cy
        if dx and dy:
            self.segments.append("l" + formatCoord((dx, dy)))
        elif dx:
            self.segments.append("h" + formatNumber(dx))
        else:
            self.segments.append("v" + formatNumber(dy))

    def _curveToOne(self, pt1, pt2, pt3):
        cx, cy = self._getCurrentPoint()
        points = [formatCoord((x - cx, y - cy)) for x, y in [pt1, pt2, pt3]]
        self.segments.append("c" + " ".join(points))

    def _qCurveToOne(self, pt1, pt2):
        cx, cy = self._getCurrentPoint()
        points = [formatCoord((x - cx, y - cy)) for x, y in [pt1, pt2]]
        self.segments.append("q" + " ".join(points))

    def _closePath(self):
        self.segments.append("Z")

    def svgPath(self):
        return " ".join(self.segments)


class SVGCanvas(Canvas):
    def __init__(self, transform):
        self.clipStack = ()
        self.currentTransform = transform
        self.elements = []

    @staticmethod
    def newPath():
        return SVGPath()

    @contextmanager
    def savedState(self):
        prevTransform = self.currentTransform
        prevClipStack = self.clipStack
        yield
        self.currentTransform = prevTransform
        self.clipStack = prevClipStack

    @contextmanager
    def compositeMode(self, compositeMode):
        yield

    def transform(self, transform):
        self.currentTransform = self.currentTransform.transform(transform)

    def clipPath(self, path):
        self.clipStack = tupleAppend(
            self.clipStack, (path.svgPath(), self.currentTransform)
        )

    def drawPathSolid(self, path, color):
        self._addElement(path.svgPath(), self.currentTransform, RGBAPaint(color), None)

    def drawPathLinearGradient(
        self, path, colorLine, pt1, pt2, extendMode, gradientTransform
    ):
        gradient = LinearGradientPaint(tuple(colorLine), pt1, pt2, extendMode)
        self._addElement(
            path.svgPath(), self.currentTransform, gradient, gradientTransform
        )

    def drawPathRadialGradient(
        self,
        path,
        colorLine,
        startCenter,
        startRadius,
        endCenter,
        endRadius,
        extendMode,
        gradientTransform,
    ):
        gradient = RadialGradientPaint(
            tuple(colorLine), startCenter, startRadius, endCenter, endRadius, extendMode
        )
        self._addElement(
            path.svgPath(), self.currentTransform, gradient, gradientTransform
        )

    def drawPathSweepGradient(
        self,
        path,
        colorLine,
        center,
        startAngle,
        endAngle,
        extendMode,
        gradientTransform,
    ):
        self.drawPathSolid(path, colorLine[0][1])

    # TODO: blendMode for PaintComposite

    def _addElement(self, fillPath, fillTransform, paint, gradientTransform):
        clipPath, clipTransform = None, None
        if self.clipStack:
            clipPath, clipTransform = self.clipStack[-1]
            if len(self.clipStack) > 1:
                # FIXME: intersect clip paths with pathops
                logger.warning(
                    "SVG canvas does not support more than two nested clip paths"
                )
        self.elements.append(
            (fillPath, fillTransform, clipPath, clipTransform, paint, gradientTransform)
        )


class RGBAPaint(tuple):
    pass


class LinearGradientPaint(NamedTuple):
    colorLine: tuple
    pt1: tuple
    pt2: tuple
    extendMode: str

    def toSVG(self, gradientID, transform):
        attrNumbers = [
            ("x1", self.pt1[0]),
            ("y1", self.pt1[1]),
            ("x2", self.pt2[0]),
            ("y2", self.pt2[1]),
        ]
        return _gradientToSVG(
            "linearGradient",
            gradientID,
            self.extendMode,
            self.colorLine,
            transform,
            attrNumbers,
        )


class RadialGradientPaint(NamedTuple):
    colorLine: tuple
    pt1: tuple
    radius1: float
    pt2: tuple
    radius2: float
    extendMode: str

    def toSVG(self, gradientID, transform):
        attrNumbers = [
            ("fx", self.pt1[0]),
            ("fy", self.pt1[1]),
            ("fr", self.radius1),
            ("cx", self.pt2[0]),
            ("cy", self.pt2[1]),
            ("r", self.radius2),
        ]
        return _gradientToSVG(
            "radialGradient",
            gradientID,
            self.extendMode,
            self.colorLine,
            transform,
            attrNumbers,
        )


def _gradientToSVG(
    gradientTag, gradientID, extendMode, colorLine, transform, attrNumbers
):
    element = ET.Element(
        gradientTag,
        id=gradientID,
        spreadMethod=_extendModeMap[extendMode],
        gradientUnits="userSpaceOnUse",
    )
    for attrName, value in attrNumbers:
        element.attrib[attrName] = formatNumber(value)
    if transform != (1, 0, 0, 1, 0, 0):
        element.attrib["gradientTransform"] = formatMatrix(transform)
    for stop, rgba in colorLine:
        stopElement = ET.SubElement(element, "stop")
        stopElement.attrib["offset"] = f"{round(stop * 100)}%"
        for attr, value in colorToSVGAttrs(rgba, "stop-color", "stop-opacity"):
            stopElement.attrib[attr] = value
    return element


class SVGSurface(Surface):
    fileExtension = ".svg"

    def __init__(self):
        self._svgElements = None

    @contextmanager
    def canvas(self, boundingBox):
        x, y, xMax, yMax = boundingBox
        width = xMax - x
        height = yMax - y
        self._viewBox = x, y, width, height
        transform = Transform(1, 0, 0, -1, 0, height + 2 * y)
        canvas = SVGCanvas(transform)
        yield canvas
        self._svgElements = canvas.elements

    def saveImage(self, path):
        with open(path, "wb") as f:
            writeSVGElements(self._svgElements, self._viewBox, f)


def writeSVGElements(elements, viewBox, stream):
    clipPaths = {}
    gradients = {}
    for fillPath, fillT, clipPath, clipT, paint, paintT in elements:
        clipKey = clipPath, clipT
        if clipPath is not None and clipKey not in clipPaths:
            clipPaths[clipKey] = f"clip_{len(clipPaths)}"
        gradientKey = paint, paintT
        if not isinstance(paint, RGBAPaint) and gradientKey not in gradients:
            gradients[gradientKey] = f"gradient_{len(gradients)}"

    root = ET.Element(
        "svg",
        width=formatNumber(viewBox[2]),
        height=formatNumber(viewBox[3]),
        preserveAspectRatio="xMinYMin slice",
        viewBox=" ".join(formatNumber(n) for n in viewBox),
        version="1.1",
        xmlns="http://www.w3.org/2000/svg",
    )

    # root.attrib["xmlns:link"] = "http://www.w3.org/1999/xlink"

    if gradients:
        defs = ET.SubElement(root, "defs")
        for (gradient, gradientTransform), gradientID in gradients.items():
            defs.append(gradient.toSVG(gradientID, gradientTransform))

    for (clipPath, clipTransform), clipID in clipPaths.items():
        clipElement = ET.SubElement(root, "clipPath", id=clipID)
        ET.SubElement(
            clipElement, "path", d=clipPath, transform=formatMatrix(clipTransform)
        )

    for fillPath, fillT, clipPath, clipT, paint, paintT in elements:
        attrs = [("d", fillPath)]
        if isinstance(paint, RGBAPaint):
            attrs += colorToSVGAttrs(paint)
        else:
            attrs.append(("fill", f"url(#{gradients[paint, paintT]})"))
        attrs.append(("transform", formatMatrix(fillT)))
        if clipPath is not None:
            clipKey = clipPath, clipTransform
            attrs.append(("clip-path", f"url(#{clipPaths[clipKey]})"))
        ET.SubElement(root, "path", dict(attrs))

    tree = ET.ElementTree(root)
    tree.write(stream, pretty_print=True, xml_declaration=True)


def formatCoord(pt):
    x, y = pt
    return "%s,%s" % (formatNumber(x), formatNumber(y))


def formatNumber(n):
    i = int(n)
    if i == n:
        return str(i)
    else:
        return str(round(n, 4))  # 4 decimals enough?


def colorToSVGAttrs(color, fillAttr="fill", opacityAttr="fill-opacity"):
    attrs = []
    opacity = 1
    if len(color) == 4:
        opacity = color[3]
        color = color[:3]
    attrs.append((fillAttr, formatColor(color)))
    if opacity != 1:
        attrs.append((opacityAttr, formatNumber(opacity)))
    return attrs


def formatColor(color):
    if not color:
        return "none"
    assert len(color) == 3
    return "#%02X%02X%02X" % tuple(int(round(c * 255)) for c in color)


def formatMatrix(t):
    assert len(t) == 6
    return "matrix(%s,%s,%s,%s,%s,%s)" % tuple(formatNumber(v) for v in t)


def tupleAppend(tpl, item):
    return tpl + (item,)



from collections import UserList
from contextlib import contextmanager
from io import BytesIO
import logging
import math
from fontTools.misc.transform import Transform, Identity
from fontTools.misc.arrayTools import unionRect
from fontTools.ttLib import TTFont
from fontTools.ttLib.tables.otTables import CompositeMode, PaintFormat
from fontTools.varLib.varStore import VarStoreInstancer
import uharfbuzz as hb


logger = logging.getLogger(__name__)


PAINT_NAMES = {v.value: k for k, v in PaintFormat.__members__.items()}
PAINT_VAR_MAPPING = {
    # Map PaintVarXxx Format to its corresponding non-var Format
    v.value: PaintFormat[k.replace("PaintVar", "Paint")]
    for k, v in PaintFormat.__members__.items()
    if k.startswith("PaintVar")
}


class BlackRendererFont:
    def __init__(self, path, *, fontNumber=0, lazy=True):
        with open(path, "rb") as f:
            fontData = f.read()
        file = BytesIO(fontData)
        self.ttFont = TTFont(file, fontNumber=fontNumber, lazy=lazy)

        self.textColor = (0, 0, 0, 1)
        self.colrV0Glyphs = {}
        self.colrV1Glyphs = {}
        self.instancer = None

        if "COLR" in self.ttFont:
            colrTable = self.ttFont["COLR"]
            if colrTable.version == 0:
                self.colrV0Glyphs = colrTable.ColorLayers
            else:  # >= 1
                # Hm, a little sad we need to use an internal static method
                self.colrV0Glyphs = colrTable._decompileColorLayersV0(colrTable.table)
                colrTable = colrTable.table
                self.colrV1Glyphs = {
                    glyph.BaseGlyph: glyph
                    for glyph in colrTable.BaseGlyphList.BaseGlyphPaintRecord
                }
                if colrTable.ClipList is None:
                    self.clipBoxes = None
                else:
                    self.clipBoxes = colrTable.ClipList.clips
                self.colrLayersV1 = colrTable.LayerList
                if colrTable.VarStore is not None:
                    self.instancer = VarStoreInstancer(
                        colrTable.VarStore, self.ttFont["fvar"].axes
                    )
                else:
                    self.instancer = None

        if "CPAL" in self.ttFont:
            self.palettes = _unpackPalettes(self.ttFont["CPAL"].palettes)
            self.currentPalette = self.palettes[0]
        else:
            self.palettes = None
            self.currentPalette = None

        if "fvar" in self.ttFont:
            self.axisTags = [a.axisTag for a in self.ttFont["fvar"].axes]
        else:
            self.axisTags = []

        self.hbFont = hb.Font(hb.Face(fontData, fontNumber))

    @property
    def unitsPerEm(self):
        return self.hbFont.face.upem

    def setLocation(self, location):
        if location is None:
            location = {}
        self.hbFont.set_variations(location)
        if self.instancer is not None:
            normalizedAxisValues = self.hbFont.get_var_coords_normalized()
            normalizedLocation = axisValuesToLocation(
                normalizedAxisValues, self.axisTags
            )
            self.instancer.setLocation(normalizedLocation)

    @property
    def glyphNames(self):
        return self.ttFont.getGlyphOrder()

    @property
    def colrV0GlyphNames(self):
        return self.colrV0Glyphs.keys()

    @property
    def colrV1GlyphNames(self):
        return self.colrV1Glyphs.keys()

    def getGlyphBounds(self, glyphName):
        if glyphName in self.colrV1Glyphs:
            bounds = self._getGlyphBounds(glyphName)
            if self.clipBoxes is not None:
                box = self.clipBoxes.get(glyphName)
                if box is not None:
                    bounds = box.xMin, box.yMin, box.xMax, box.yMax
        elif glyphName in self.colrV0Glyphs:
            # For COLRv0, we take the union of all layer bounds
            bounds = None
            for layer in self.colrV0Glyphs[glyphName]:
                layerBounds = self._getGlyphBounds(layer.name)
                if bounds is None:
                    bounds = layerBounds
                else:
                    bounds = unionRect(layerBounds, bounds)
        else:
            bounds = self._getGlyphBounds(glyphName)
        return bounds

    def drawGlyph(self, glyphName, canvas, *, palette=None, textColor=(0, 0, 0, 1)):
        if palette is None and self.palettes:
            palette = self.palettes[0]
        self.currentPalette = palette
        self.textColor = textColor

        glyph = self.colrV1Glyphs.get(glyphName)
        if glyph is not None:
            self.currentTransform = Identity
            self.currentPath = None
            self._drawGlyphCOLRv1(glyph, canvas)
            return
        glyph = self.colrV0Glyphs.get(glyphName)
        if glyph is not None:
            self._drawGlyphCOLRv0(glyph, canvas)
            return
        else:
            self._drawGlyphNoColor(glyphName, canvas)

    def _drawGlyphNoColor(self, glyphName, canvas):
        path = canvas.newPath()
        self._drawGlyphOutline(glyphName, path)
        canvas.drawPathSolid(path, self.textColor)

    def _drawGlyphCOLRv0(self, layers, canvas):
        for layer in layers:
            path = canvas.newPath()
            self._drawGlyphOutline(layer.name, path)
            canvas.drawPathSolid(path, self._getColor(layer.colorID, 1))

    def _drawGlyphCOLRv1(self, glyph, canvas):
        self._drawPaint(glyph.Paint, canvas)

    # COLRv1 Paint dispatch

    def _drawPaint(self, paint, canvas):
        nonVarFormat = PAINT_VAR_MAPPING.get(paint.Format)
        if nonVarFormat is None:
            # "regular" Paint
            paintName = PAINT_NAMES.get(paint.Format)
            if paintName is None:
                logger.warning(f"Ignoring unknown COLRv1 Paint format: {paint.Format}")
                return
        else:
            # PaintVar -- we map to its non-var counterpart and use a wrapper
            # that takes care of instantiating values
            paintName = PAINT_NAMES[nonVarFormat]
            paint = PaintVarWrapper(paint, self.instancer)
        drawHandler = getattr(self, "_draw" + paintName)
        drawHandler(paint, canvas)

    def _drawPaintColrLayers(self, paint, canvas):
        n = paint.NumLayers
        s = paint.FirstLayerIndex
        with self._ensureClipAndPushPath(canvas, None):
            for i in range(s, s + n):
                with self._savedTransform():
                    self._drawPaint(self.colrLayersV1.Paint[i], canvas)

    def _drawPaintSolid(self, paint, canvas):
        color = self._getColor(paint.PaletteIndex, paint.Alpha)
        canvas.drawPathSolid(self.currentPath, color)

    def _drawPaintLinearGradient(self, paint, canvas):
        minStop, maxStop, colorLine = self._readColorLine(paint.ColorLine)
        pt1, pt2 = _reduceThreeAnchorsToTwo(paint)
        pt1, pt2 = (
            _interpolatePoints(pt1, pt2, minStop),
            _interpolatePoints(pt1, pt2, maxStop),
        )
        canvas.drawPathLinearGradient(
            self.currentPath,
            colorLine,
            pt1,
            pt2,
            paint.ColorLine.Extend,
            self.currentTransform,
        )

    def _drawPaintRadialGradient(self, paint, canvas):
        minStop, maxStop, colorLine = self._readColorLine(paint.ColorLine)
        startCenter = (paint.x0, paint.y0)
        endCenter = (paint.x1, paint.y1)
        startCenter, endCenter = (
            _interpolatePoints(startCenter, endCenter, minStop),
            _interpolatePoints(startCenter, endCenter, maxStop),
        )
        startRadius = _interpolate(paint.r0, paint.r1, minStop)
        endRadius = _interpolate(paint.r0, paint.r1, maxStop)
        canvas.drawPathRadialGradient(
            self.currentPath,
            colorLine,
            startCenter,
            startRadius,
            endCenter,
            endRadius,
            paint.ColorLine.Extend,
            self.currentTransform,
        )

    def _drawPaintSweepGradient(self, paint, canvas):
        minStop, maxStop, colorLine = self._readColorLine(paint.ColorLine)
        center = paint.centerX, paint.centerY
        startAngle = _interpolate(paint.startAngle, paint.endAngle, minStop)
        endAngle = _interpolate(paint.startAngle, paint.endAngle, maxStop)
        canvas.drawPathSweepGradient(
            self.currentPath,
            colorLine,
            center,
            startAngle,
            endAngle,
            paint.ColorLine.Extend,
            self.currentTransform,
        )

    def _drawPaintGlyph(self, paint, canvas):
        path = canvas.newPath()
        # paint.Glyph must not be a COLR glyph
        self._drawGlyphOutline(paint.Glyph, path)
        with self._ensureClipAndPushPath(canvas, path):
            self._drawPaint(paint.Paint, canvas)

    def _drawPaintColrGlyph(self, paint, canvas):
        with self._ensureClipAndPushPath(canvas, None):
            self._drawGlyphCOLRv1(self.colrV1Glyphs[paint.Glyph], canvas)

    def _drawPaintTransform(self, paint, canvas):
        t = paint.Transform
        transform = (t.xx, t.yx, t.xy, t.yy, t.dx, t.dy)
        self._applyTransform(transform, paint.Paint, canvas)

    def _drawPaintTranslate(self, paint, canvas):
        transform = (1, 0, 0, 1, paint.dx, paint.dy)
        self._applyTransform(transform, paint.Paint, canvas)

    def _drawPaintRotate(self, paint, canvas):
        transform = Transform()
        transform = transform.rotate(math.radians(paint.angle))
        self._applyTransform(transform, paint.Paint, canvas)

    def _drawPaintRotateAroundCenter(self, paint, canvas):
        transform = Transform()
        transform = transform.translate(paint.centerX, paint.centerY)
        transform = transform.rotate(math.radians(paint.angle))
        transform = transform.translate(-paint.centerX, -paint.centerY)
        self._applyTransform(transform, paint.Paint, canvas)

    def _drawPaintSkew(self, paint, canvas):
        transform = Transform()
        transform = transform.skew(
            math.radians(paint.xSkewAngle), math.radians(paint.ySkewAngle)
        )
        self._applyTransform(transform, paint.Paint, canvas)

    def _drawPaintSkewAroundCenter(self, paint, canvas):
        transform = Transform()
        transform = transform.translate(paint.centerX, paint.centerY)
        transform = transform.skew(
            math.radians(paint.xSkewAngle), math.radians(paint.ySkewAngle)
        )
        transform = transform.translate(-paint.centerX, -paint.centerY)
        self._applyTransform(transform, paint.Paint, canvas)

    def _drawPaintScale(self, paint, canvas):
        transform = Transform()
        transform = transform.scale(paint.scaleX, paint.scaleY)
        self._applyTransform(transform, paint.Paint, canvas)

    def _drawPaintScaleAroundCenter(self, paint, canvas):
        transform = Transform()
        transform = transform.translate(paint.centerX, paint.centerY)
        transform = transform.scale(paint.scaleX, paint.scaleY)
        transform = transform.translate(-paint.centerX, -paint.centerY)
        self._applyTransform(transform, paint.Paint, canvas)

    def _drawPaintScaleUniform(self, paint, canvas):
        transform = Transform()
        transform = transform.scale(paint.scale, paint.scale)
        self._applyTransform(transform, paint.Paint, canvas)

    def _drawPaintScaleUniformAroundCenter(self, paint, canvas):
        transform = Transform()
        transform = transform.translate(paint.centerX, paint.centerY)
        transform = transform.scale(paint.scale, paint.scale)
        transform = transform.translate(-paint.centerX, -paint.centerY)
        self._applyTransform(transform, paint.Paint, canvas)

    def _drawPaintComposite(self, paint, canvas):
        with self._ensureClipAndPushPath(canvas, None):
            with canvas.compositeMode(CompositeMode.SRC_OVER):
                with self._savedTransform():
                    self._drawPaint(paint.BackdropPaint, canvas)
                with canvas.compositeMode(paint.CompositeMode):
                    with self._savedTransform():
                        self._drawPaint(paint.SourcePaint, canvas)

    def _drawPaintLocation(self, paint, canvas):
        # https://github.com/googlefonts/colr-gradients-spec/issues/277
        numAxes = len(self.axisTags)
        location = {
            self.axisTags[coord.AxisIndex]: coord.AxisValue
            for coord in paint.Coordinate
            if coord.AxisIndex < numAxes
        }
        with self._pushNormalizedLocation(location):
            self._drawPaint(paint.Paint, canvas)

    # Utils

    @contextmanager
    def _pushNormalizedLocation(self, location):
        savedAxisValues = self.hbFont.get_var_coords_normalized()
        tmpAxisValues = list(savedAxisValues)
        if len(tmpAxisValues) < len(self.axisTags):
            # pad with zeros
            tmpAxisValues.extend([0] * (len(self.axisTags) - len(tmpAxisValues)))
        for axisIndex, axisTag in enumerate(self.axisTags):
            axisValue = location.get(axisTag)
            if axisValue is not None:
                tmpAxisValues[axisIndex] = axisValue
        tmpLocation = axisValuesToLocation(tmpAxisValues, self.axisTags)

        self.hbFont.set_var_coords_normalized(tmpAxisValues)
        if self.instancer is not None:
            savedLocation = self.instancer.location
            # FIXME: calling setLocation loses the internal instancer._scalars cache;
            # perhaps a pushing a *new* instancer and reverting to the old one is
            # faster, but this is currently not possible due to PaintVarXxx referencing
            # self.instancer, too.
            self.instancer.setLocation(tmpLocation)
            yield
            self.instancer.setLocation(savedLocation)
        else:
            yield
        self.hbFont.set_var_coords_normalized(savedAxisValues)

    @contextmanager
    def _ensureClipAndPushPath(self, canvas, path):
        currentPath = self.currentPath
        currentTransform = self.currentTransform
        self.currentPath = path
        self.currentTransform = Identity
        with canvas.savedState():
            if currentTransform != Identity:
                canvas.transform(currentTransform)
            if currentPath is not None:
                canvas.clipPath(currentPath)
            yield
        self.currentPath = currentPath
        self.currentTransform = currentTransform

    @contextmanager
    def _savedTransform(self):
        savedTransform = self.currentTransform
        yield
        self.currentTransform = savedTransform

    def _applyTransform(self, transform, paint, canvas):
        self.currentTransform = self.currentTransform.transform(transform)
        self._drawPaint(paint, canvas)

    def _drawGlyphOutline(self, glyphName, path):
        gid = self.ttFont.getGlyphID(glyphName)
        self.hbFont.draw_glyph_with_pen(gid, path)

    def _getGlyphBounds(self, glyphName):
        gid = self.ttFont.getGlyphID(glyphName)
        x, y, w, h = self.hbFont.get_glyph_extents(gid)
        # convert from HB's x/y_bearing + extents to xMin, yMin, xMax, yMax
        y += h
        h = -h
        w += x
        h += y
        return x, y, w, h

    def _getColor(self, colorIndex, alpha):
        if (
            colorIndex == 0xFFFF
            or self.currentPalette is None
            or colorIndex >= len(self.currentPalette)
        ):
            r, g, b, a = self.textColor
        else:
            r, g, b, a = self.currentPalette[colorIndex]
        a *= alpha
        return r, g, b, a

    def _readColorLine(self, colorLineTable):
        return _normalizeColorLine(
            [
                (cs.StopOffset, self._getColor(cs.PaletteIndex, cs.Alpha))
                for cs in colorLineTable.ColorStop
            ]
        )


def _reduceThreeAnchorsToTwo(p):
    # FIXME: make sure the 3 points are not in degenerate position [see COLRv1 spec].
    x02 = p.x2 - p.x0
    y02 = p.y2 - p.y0
    x01 = p.x1 - p.x0
    y01 = p.y1 - p.y0
    squaredNorm02 = x02 * x02 + y02 * y02
    k = (x01 * x02 + y01 * y02) / squaredNorm02
    x = p.x1 - k * x02
    y = p.y1 - k * y02
    return ((p.x0, p.y0), (x, y))


def _normalizeColorLine(colorLine):
    stops = [stopOffset for stopOffset, color in colorLine]
    minStop = min(stops)
    maxStop = max(stops)
    if minStop != maxStop:
        stopExtent = maxStop - minStop
        colorLine = [
            ((stopOffset - minStop) / stopExtent, color)
            for stopOffset, color in colorLine
        ]
    else:
        # Degenerate case. minStop and maxStop are used to reposition
        # the gradients parameters (points, radii, angles), through
        # interpolation. By setting minStop and maxStop to 0 and 1,
        # we at least won't mess up the parameters.
        # FIXME: should the gradient simply not be drawn in this case?
        # 1. If there's only one color stop, we can just draw a solid
        # color.
        # 2. If there are > 1 color stops all at the same color stop
        # offset, things get interesting: everything "to the left"
        # should get color[0], everything "to the right" should get
        # color[-1]. If there are more than 2 color stops, all but the
        # first and last should be ignored.
        minStop, maxStop = 0, 1
    return minStop, maxStop, colorLine


def _interpolate(v1, v2, f):
    return v1 + f * (v2 - v1)


def _interpolatePoints(pt1, pt2, f):
    x1, y1 = pt1
    x2, y2 = pt2
    return (x1 + f * (x2 - x1), y1 + f * (y2 - y1))


def _unpackPalettes(palettes):
    return [
        [(c.red / 255, c.green / 255, c.blue / 255, c.alpha / 255) for c in p]
        for p in palettes
    ]


def axisValuesToLocation(normalizedAxisValues, axisTags):
    return {
        axisTag: axisValue for axisTag, axisValue in zip(axisTags, normalizedAxisValues)
    }


# _conversionFactors = {
#     VarF2Dot14: 1 / (1 << 14),
#     VarFixed: 1 / (1 << 16),
# }


class PaintVarWrapper:
    def __init__(self, wrappedPaint, instancer):
        assert not isinstance(wrappedPaint, PaintVarWrapper)
        self._wrappedPaint = wrappedPaint
        self._instancer = instancer

    def __repr__(self):
        return f"PaintVarWrapper({self._wrappedPaint!r})"

    def __getattr__(self, attrName):
        value = getattr(self._wrappedPaint, attrName)
        raise NotImplementedError("This code is currently not working")
        # if isinstance(value, VariableValue):
        #     if value.varIdx != 0xFFFFFFFF:
        #         factor = _conversionFactors.get(
        #             type(self._wrappedPaint.getConverterByName(attrName)), 1
        #         )
        #         value = value.value + self._instancer[value.varIdx] * factor
        #     else:
        #         value = value.value
        # elif type(value).__name__.startswith("Var"):
        #     value = PaintVarWrapper(value, self._instancer)
        # elif isinstance(value, (list, UserList)):
        #     value = [PaintVarWrapper(item, self._instancer) for item in value]
        return value



from functools import singledispatch
from fontTools.ttLib.tables.otTables import ColorLine, Paint
from .font import PAINT_NAMES


def dumpCOLRv1Glyph(font, glyphName):
    print(glyphName)
    glyph = font.colrV1Glyphs[glyphName]
    d = unpackObject(glyph.Paint, font)
    printObject(d, 0)


simpleTypes = (int, float, str, tuple)


@singledispatch
def unpackObject(obj, font):
    d = {}
    for n, v in obj.__dict__.items():
        if hasattr(v, "__dict__"):
            v = unpackObject(v, font)
        d[n] = v
    return d


@unpackObject.register
def unpackPaint(paint: Paint, font):
    paintName = PAINT_NAMES[paint.Format]
    d = {"#": paintName}
    if paintName == "PaintColrLayers":
        n = paint.NumLayers
        s = paint.FirstLayerIndex
        layers = [
            unpackObject(font.colrLayersV1.Paint[i], font) for i in range(s, s + n)
        ]
        d["Layers"] = layers
    else:
        for n, v in paint.__dict__.items():
            if not isinstance(v, simpleTypes):
                v = unpackObject(v, font)
            d[n] = v
    return d


@unpackObject.register
def unpackColorLine(colorLine: ColorLine, font):
    return [
        (
            round(cs.StopOffset, 3),
            color255(font._getColor(cs.Color.PaletteIndex, cs.Color.Alpha)),
        )
        for cs in colorLine.ColorStop
    ]


def color255(color):
    return tuple(round(ch * 255) for ch in color)


@singledispatch
def printObject(obj, level, prefix="  "):
    print("    " * level + prefix + reprItem(obj))


@printObject.register
def printDict(dct: dict, level, prefix="  "):
    for k, v in dct.items():
        if isinstance(v, simpleTypes):
            print("    " * level + prefix + str(k), reprItem(v))
        else:
            print("    " * level + prefix + str(k))
            printObject(v, level + 1)
        prefix = "  "


@printObject.register
def printList(lst: list, level):
    for item in lst:
        printObject(item, level, "- ")


def reprItem(item):
    if isinstance(item, float):
        return str(round(item, 3))
    else:
        return str(item)



import pathlib
import pytest
from fontTools.misc.transform import Identity
from fontTools.ttLib.tables.otTables import CompositeMode, ExtendMode
from blackrenderer.backends import getSurfaceClass
from compareImages import compareImages


testDir = pathlib.Path(__file__).resolve().parent
dataDir = testDir / "data"
expectedOutputDir = testDir / "expectedOutput"
tmpOutputDir = testDir / "tmpOutput"
if not tmpOutputDir.exists():
    tmpOutputDir.mkdir()


backends = [
    (name, getSurfaceClass(name)) for name in ["cairo", "coregraphics", "skia", "svg"]
]
backends = [(name, surface) for name, surface in backends if surface is not None]


test_colorStopValues = [
    (0, 1),
]

test_extendModes = [ExtendMode.PAD, ExtendMode.REPEAT, ExtendMode.REFLECT]


@pytest.mark.parametrize("stopOffsets", test_colorStopValues)
@pytest.mark.parametrize("extend", test_extendModes)
@pytest.mark.parametrize("backendName, surfaceClass", backends)
def test_colorStops(backendName, surfaceClass, stopOffsets, extend):
    point1 = (200, 0)
    point2 = (400, 0)
    color1 = (1, 0, 0, 1)
    color2 = (0, 0, 1, 1)
    stop1, stop2 = stopOffsets
    colorLine = [(stop1, color1), (stop2, color2)]

    surface = surfaceClass()
    with surface.canvas((0, 0, 600, 100)) as canvas:
        canvas.drawRectLinearGradient(
            (0, 0, 600, 100), colorLine, point1, point2, extend, Identity
        )
        for pos in [200, 400]:
            canvas.drawRectSolid((pos, 0, 1, 100), (0, 0, 0, 1))

    ext = surface.fileExtension
    stopsString = "_".join(str(s) for s in stopOffsets)
    fileName = f"colorStops_{extend.name}_{stopsString}_{backendName}{ext}"
    expectedPath = expectedOutputDir / fileName
    outputPath = tmpOutputDir / fileName
    surface.saveImage(outputPath)
    assert expectedPath.read_bytes() == outputPath.read_bytes()


@pytest.mark.parametrize("extend", test_extendModes)
@pytest.mark.parametrize("backendName, surfaceClass", backends)
def test_sweepGradient(backendName, surfaceClass, extend):
    H, W = 400, 400
    center = (H / 2, W / 2)
    startAngle = 45
    endAngle = 315
    colors = [
        (1, 0, 0, 1),
        (0, 1, 0, 1),
        (1, 1, 0, 1),
        (1, 0.5, 1, 1),
        (0, 0, 1, 1),
    ]
    stopOffsets = [0, 0.5, 0.5, 0.6, 1]
    colorLine = list(zip(stopOffsets, colors))

    surface = surfaceClass()
    with surface.canvas((0, 0, H, W)) as canvas:
        canvas.drawRectSweepGradient(
            (0, 0, H, W), colorLine, center, startAngle, endAngle, extend, Identity
        )

    ext = surface.fileExtension
    stopsString = "_".join(str(s) for s in stopOffsets)
    fileName = f"sweepGradient_{extend.name}_{stopsString}_{backendName}{ext}"
    expectedPath = expectedOutputDir / fileName
    outputPath = tmpOutputDir / fileName
    surface.saveImage(outputPath)
    assert expectedPath.read_bytes() == outputPath.read_bytes()


test_compositeModes = [
    CompositeMode.CLEAR,
    CompositeMode.SRC,
    CompositeMode.DEST,
    CompositeMode.SRC_OVER,
    CompositeMode.DEST_OVER,
    CompositeMode.SRC_IN,
    CompositeMode.DEST_IN,
    CompositeMode.SRC_OUT,
    CompositeMode.DEST_OUT,
    CompositeMode.SRC_ATOP,
    CompositeMode.DEST_ATOP,
    CompositeMode.XOR,
    CompositeMode.PLUS,
    CompositeMode.SCREEN,
    CompositeMode.OVERLAY,
    CompositeMode.DARKEN,
    CompositeMode.LIGHTEN,
    CompositeMode.COLOR_DODGE,
    CompositeMode.COLOR_BURN,
    CompositeMode.HARD_LIGHT,
    CompositeMode.SOFT_LIGHT,
    CompositeMode.DIFFERENCE,
    CompositeMode.EXCLUSION,
    CompositeMode.MULTIPLY,
    CompositeMode.HSL_HUE,
    CompositeMode.HSL_SATURATION,
    CompositeMode.HSL_COLOR,
    CompositeMode.HSL_LUMINOSITY,
]


@pytest.mark.parametrize("compositeMode", test_compositeModes)
@pytest.mark.parametrize("backendName, surfaceClass", backends)
def test_compositeMode(backendName, surfaceClass, compositeMode):
    H, W = 400, 400
    surface = surfaceClass()
    with surface.canvas((0, 0, H, W)) as canvas:
        canvas.drawRectSolid((50, 50, 200, 200), (1, 0.2, 0, 1))
        with canvas.compositeMode(compositeMode):
            canvas.drawRectSolid((150, 150, 200, 200), (0, 0.2, 1, 1))
    ext = surface.fileExtension
    compositeModeName = compositeMode.name.replace("_", "")  # nicer file sorting
    fileName = f"compositeMode_{compositeModeName}_{backendName}{ext}"
    expectedPath = expectedOutputDir / fileName
    outputPath = tmpOutputDir / fileName
    surface.saveImage(outputPath)
    diff = compareImages(expectedPath, outputPath)
    assert diff < 0.00013, diff



import io
import re
import sys
from PIL import Image, ImageChops


def compareImages(path1, path2):
    """Compare two image files and return a number representing how similar they are.
    A value of 0 means that the images are identical, a value of 1 means they are maximally
    different or not comparable (for example, when their dimensions differ).
    """
    assert path1.suffix == path2.suffix
    suffix = path1.suffix.lower()
    if suffix == ".svg":
        svgEqual = compareSVG(path1, path2)
        if svgEqual:
            return 0
        else:
            return 1
    elif suffix == ".pdf":
        pdfEqual, im1, im2 = comparePDF(path1, path2)
        if pdfEqual:
            return 0
        if im1 is None or im2 is None:
            return 1
    else:
        im1 = Image.open(path1)
        im2 = Image.open(path2)

    if im1.size != im2.size:
        # Dimensions differ, can't compare further
        return 1

    if im1 == im2:
        # Image data is identical (I checked PIL's Image.__eq__ method: it's solid)
        return 0

    # Get the difference between the images
    diff = ImageChops.difference(im1, im2)

    # We'll calculate the average difference based on the histogram provided by PIL
    hist = diff.histogram()
    assert (
        len(hist) == 4 * 256
    )  # Assuming 4x8-bit RGBA for now. TODO: make this work for L and RGB modes
    # Sum the histograms of each channel
    summedHist = [
        sum(hist[pixelValue + ch * 256] for ch in range(4)) for pixelValue in range(256)
    ]

    assert len(summedHist) == 256
    assert sum(hist) == sum(summedHist)
    # Calculate the average of the difference
    # First add all pixel values together
    totalSum = sum(summedHist[pixelValue] * pixelValue for pixelValue in range(256))
    # Then divide by the total number of channel values
    average = totalSum / sum(summedHist)
    # Scale pixel value range from 0-255 to 0-1
    average = average / 255
    assert 0.0 <= average <= 1.0
    return average


def compareFiles(path1, path2):
    return path1.read_bytes() != path2.read_bytes()


_svgIgnore = [
    (rb'"surface\d+"', b'"surface***"'),
    (rb'"#surface\d+"', b'"#surface***"'),
]


def compareSVG(path1, path2):
    data1 = path1.read_bytes()
    data2 = path2.read_bytes()
    return _filterData(data1, _svgIgnore) == _filterData(data2, _svgIgnore)


_pdfIgnore = [
    (rb"/CreationDate \([^)]+\)", b"/CreationDate (00000000)"),
    (rb"0000000000 65535 f[^t]+trailer", b"******"),
    (rb"startxref[^%]+%%EOF", b"******"),
]


def comparePDF(path1, path2):
    data1 = path1.read_bytes()
    data2 = path2.read_bytes()
    if _filterData(data1, _pdfIgnore) == _filterData(data2, _pdfIgnore):
        return True, None, None
    if sys.platform == "darwin":
        im1 = macRenderPDF(data1)
        im2 = macRenderPDF(data2)
        return None, im1, im2
    else:
        return False, None, None


def _filterData(data, ignorePatterns):
    for pat, repl in ignorePatterns:
        data = re.sub(pat, repl, data)
    return data


def macRenderPDF(data):
    import AppKit

    pdf = AppKit.PDFDocument.alloc().initWithData_(data)
    if pdf.pageCount() != 1:
        return None
    page = pdf.pageAtIndex_(0)
    image = AppKit.NSImage.alloc().initWithData_(page.dataRepresentation())
    return Image.open(io.BytesIO(image.TIFFRepresentation()))



import os
import pathlib
import subprocess
import pytest


dataDir = pathlib.Path(__file__).resolve().parent / "data"


testData = [
    ("ABC", dataDir / "MutatorSans.ttf"),
    ("\u2693\u2694", dataDir / "noto-glyf_colr_1.ttf"),
    ("\u3299", dataDir / "TwemojiMozilla.subset.default.3299.ttf"),
]


@pytest.mark.parametrize("testString, fontPath", testData)
@pytest.mark.parametrize("outputFormat", [".png", ".svg"])
def test_mainprog(tmpdir, testString, fontPath, outputFormat):
    outputPath = os.path.join(tmpdir, "test" + outputFormat)
    args = [
        "blackrenderer",
        os.fspath(fontPath),
        testString,
        outputPath,
        "--font-size",
        "50",
    ]
    subprocess.check_output(args)
    assert os.path.isfile(outputPath)


expectedSVGOutput = """\
<?xml version='1.0' encoding='ASCII'?>
<svg width="42" height="75" preserveAspectRatio="xMinYMin slice" viewBox="-17 -20 42 75" version="1.1" xmlns="http://www.w3.org/2000/svg">
  <path d="M60,0 v700 h40 v-700 h-40 Z" fill="#000000" transform="matrix(0.05,0,0,-0.05,0,35)"/>
</svg>
"""


def test_mainprog_svg_stdout():
    args = [
        "blackrenderer",
        os.fspath(dataDir / "MutatorSans.ttf"),
        "I",
        "-",
        "--font-size",
        "50",
    ]
    output = subprocess.check_output(args, shell=False, encoding="ascii")
    assert expectedSVGOutput.splitlines() == output.splitlines()



from contextlib import contextmanager
import os
from math import sqrt
from fontTools.pens.basePen import BasePen
from fontTools.pens.recordingPen import RecordingPen
from fontTools.ttLib.tables.otTables import CompositeMode, ExtendMode
import cairo
from .base import Canvas, Surface
from .sweepGradient import buildSweepGradientPatches


_compositeModeMap = {
    CompositeMode.CLEAR: cairo.OPERATOR_CLEAR,
    CompositeMode.SRC: cairo.OPERATOR_SOURCE,
    CompositeMode.DEST: cairo.OPERATOR_DEST,
    CompositeMode.SRC_OVER: cairo.OPERATOR_OVER,
    CompositeMode.DEST_OVER: cairo.OPERATOR_DEST_OVER,
    CompositeMode.SRC_IN: cairo.OPERATOR_IN,
    CompositeMode.DEST_IN: cairo.OPERATOR_DEST_IN,
    CompositeMode.SRC_OUT: cairo.OPERATOR_OUT,
    CompositeMode.DEST_OUT: cairo.OPERATOR_DEST_OUT,
    CompositeMode.SRC_ATOP: cairo.OPERATOR_ATOP,
    CompositeMode.DEST_ATOP: cairo.OPERATOR_DEST_ATOP,
    CompositeMode.XOR: cairo.OPERATOR_XOR,
    CompositeMode.PLUS: cairo.OPERATOR_ADD,
    CompositeMode.SCREEN: cairo.OPERATOR_SCREEN,
    CompositeMode.OVERLAY: cairo.OPERATOR_OVERLAY,
    CompositeMode.DARKEN: cairo.OPERATOR_DARKEN,
    CompositeMode.LIGHTEN: cairo.OPERATOR_LIGHTEN,
    CompositeMode.COLOR_DODGE: cairo.OPERATOR_COLOR_DODGE,
    CompositeMode.COLOR_BURN: cairo.OPERATOR_COLOR_BURN,
    CompositeMode.HARD_LIGHT: cairo.OPERATOR_HARD_LIGHT,
    CompositeMode.SOFT_LIGHT: cairo.OPERATOR_SOFT_LIGHT,
    CompositeMode.DIFFERENCE: cairo.OPERATOR_DIFFERENCE,
    CompositeMode.EXCLUSION: cairo.OPERATOR_EXCLUSION,
    CompositeMode.MULTIPLY: cairo.OPERATOR_MULTIPLY,
    CompositeMode.HSL_HUE: cairo.OPERATOR_HSL_HUE,
    CompositeMode.HSL_SATURATION: cairo.OPERATOR_HSL_SATURATION,
    CompositeMode.HSL_COLOR: cairo.OPERATOR_HSL_COLOR,
    CompositeMode.HSL_LUMINOSITY: cairo.OPERATOR_HSL_LUMINOSITY,
}


_extendModeMap = {
    ExtendMode.PAD: cairo.Extend.PAD,
    ExtendMode.REPEAT: cairo.Extend.REPEAT,
    ExtendMode.REFLECT: cairo.Extend.REFLECT,
}


class CairoPen(BasePen):
    def __init__(self, context):
        super().__init__(None)
        self.context = context

    def _moveTo(self, pt):
        self.context.move_to(*pt)

    def _lineTo(self, pt):
        self.context.line_to(*pt)

    def _curveToOne(self, pt1, pt2, pt3):
        self.context.curve_to(*pt1, *pt2, *pt3)

    def _closePath(self):
        self.context.close_path()


class CairoCanvas(Canvas):
    def __init__(self, context):
        self.context = context
        self._pen = CairoPen(context)

    @staticmethod
    def newPath():
        return RecordingPen()

    @contextmanager
    def savedState(self):
        self.context.save()
        yield
        self.context.restore()

    @contextmanager
    def compositeMode(self, compositeMode):
        self.context.push_group()
        yield
        self.context.pop_group_to_source()
        self.context.set_operator(_compositeModeMap[compositeMode])
        self.context.paint()

    def transform(self, transform):
        m = cairo.Matrix()
        m.xx, m.yx, m.xy, m.yy, m.x0, m.y0 = transform
        self.context.transform(m)

    def clipPath(self, path):
        self.context.new_path()
        path.replay(self._pen)
        self.context.clip()

    def drawPathSolid(self, path, color):
        self.context.set_source_rgba(*color)
        self.context.new_path()
        path.replay(self._pen)
        self.context.fill()

    def drawPathLinearGradient(
        self, path, colorLine, pt1, pt2, extendMode, gradientTransform
    ):
        gr = cairo.LinearGradient(pt1[0], pt1[1], pt2[0], pt2[1])
        gr.set_extend(_extendModeMap[extendMode])
        for stop, color in colorLine:
            gr.add_color_stop_rgba(stop, *color)
        self._drawGradient(path, gr, gradientTransform)

    def drawPathRadialGradient(
        self,
        path,
        colorLine,
        startCenter,
        startRadius,
        endCenter,
        endRadius,
        extendMode,
        gradientTransform,
    ):
        gr = cairo.RadialGradient(
            startCenter[0],
            startCenter[1],
            startRadius,
            endCenter[0],
            endCenter[1],
            endRadius,
        )
        gr.set_extend(_extendModeMap[extendMode])
        for stop, color in colorLine:
            gr.add_color_stop_rgba(stop, *color)
        self._drawGradient(path, gr, gradientTransform)

    def drawPathSweepGradient(
        self,
        path,
        colorLine,
        center,
        startAngle,
        endAngle,
        extendMode,
        gradientTransform,
    ):
        self.context.save()
        self.context.new_path()
        path.replay(self._pen)
        self.context.clip()
        self.transform(gradientTransform)
        # alloc the mesh pattern
        pat = cairo.MeshPattern()
        # find current path' extent
        x1, y1, x2, y2 = self.context.clip_extents()
        maxX = max(d * d for d in (x1 - center[0], x2 - center[0]))
        maxY = max(d * d for d in (y1 - center[1], y2 - center[1]))
        R = sqrt(maxX + maxY)
        patches = buildSweepGradientPatches(
            colorLine, center, R, startAngle, endAngle, useGouraudShading=False
        )
        for (P0, color0), C0, C1, (P1, color1) in patches:
            # draw patch
            pat.begin_patch()
            pat.move_to(center[0], center[1])
            pat.line_to(P0[0], P0[1])
            pat.curve_to(C0[0], C0[1], C1[0], C1[1], P1[0], P1[1])
            pat.line_to(center[0], center[1])
            pat.set_corner_color_rgba(0, *color0)
            pat.set_corner_color_rgba(1, *color0)
            pat.set_corner_color_rgba(2, *color1)
            pat.set_corner_color_rgba(3, *color1)
            pat.end_patch()
        self.context.set_source(pat)
        self.context.paint()
        self.context.restore()

    # TODO: blendMode for PaintComposite)

    def _drawGradient(self, path, gradient, gradientTransform):
        self.context.new_path()
        path.replay(self._pen)
        self.context.save()
        self.transform(gradientTransform)
        self.context.set_source(gradient)
        self.context.fill()
        self.context.restore()


class CairoPixelSurface(Surface):
    fileExtension = ".png"

    def __init__(self):
        self._surfaces = []

    @contextmanager
    def canvas(self, boundingBox):
        x, y, xMax, yMax = boundingBox
        width = xMax - x
        height = yMax - y
        surface = self._setupCairoSurface(width, height)
        self._surfaces.append((surface, (width, height)))
        context = cairo.Context(surface)
        context.translate(-x, height + y)
        context.scale(1, -1)
        yield CairoCanvas(context)

    def _setupCairoSurface(self, width, height):
        return cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)

    def saveImage(self, path):
        surface, _ = self._surfaces[-1]
        surface.flush()
        surface.write_to_png(os.fspath(path))
        surface.finish()


class CairoPDFSurface(CairoPixelSurface):
    fileExtension = ".pdf"

    def _setupCairoSurface(self, width, height):
        return cairo.RecordingSurface(cairo.CONTENT_COLOR_ALPHA, (0, 0, width, height))

    def saveImage(self, path):
        _, (width, height) = self._surfaces[0]
        pdfSurface = cairo.PDFSurface(path, width, height)
        pdfContext = None
        for surface, (width, height) in self._surfaces:
            pdfSurface.set_size(width, height)
            if pdfContext is None:
                # It's important to call the first set_size() *before*
                # the context is created, or we'll get an additional
                # empty page
                pdfContext = cairo.Context(pdfSurface)
            pdfContext.set_source_surface(surface, 0.0, 0.0)
            pdfContext.paint()
            pdfContext.show_page()
        pdfSurface.flush()


class CairoSVGSurface(CairoPDFSurface):
    fileExtension = ".svg"

    def saveImage(self, path):
        surface, (width, height) = self._surfaces[-1]
        svgSurface = cairo.SVGSurface(path, width, height)
        pdfContext = cairo.Context(svgSurface)
        pdfContext.set_source_surface(surface, 0.0, 0.0)
        pdfContext.paint()
        svgSurface.flush()



import pathlib
from blackrenderer.font import BlackRendererFont


testDir = pathlib.Path(__file__).resolve().parent
testFont1 = testDir / "data" / "noto-glyf_colr_1.ttf"


def test_font():
    font = BlackRendererFont(testFont1)
    assert len(font.glyphNames) > len(font.colrV0GlyphNames)
    assert len(font.glyphNames) > len(font.colrV1GlyphNames)



# Copyright 2021 Black Foundry. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="blackrenderer",
    use_scm_version={"write_to": "Lib/blackrenderer/_version.py"},
    description=(
        "A Python-based renderer for OpenType COLRv1 fonts, with multiple backends."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Black Foundry, Just van Rossum, Samuel Hornus",
    author_email="justvanrossum@gmail.com",
    license="Apache Software License 2.0",
    url="https://github.com/BlackFoundryCom/black-renderer",
    python_requires=">=3.7",
    package_dir={"": "Lib"},
    packages=find_packages("Lib"),
    install_requires=[
        "fonttools >= 4.17.0",
        "uharfbuzz >= 0.16.0",
    ],
    extras_require={
        "skia": ["skia-python", "numpy"],
        "cairo": ["pycairo"],
        "cg": ["pyobjc; sys_platform == 'darwin'"],
    },
    setup_requires=["setuptools_scm"],
    entry_points={
        "console_scripts": [
            "blackrenderer=blackrenderer.__main__:main",
        ],
    },
)



from typing import NamedTuple
import os
from fontTools.misc.arrayTools import (
    scaleRect,
    offsetRect,
    intRect,
    unionRect,
    insetRect,
)
import uharfbuzz as hb
from .font import BlackRendererFont
from .backends import getSurfaceClass


class BackendUnavailableError(Exception):
    pass


def renderText(
    fontPath,
    textString,
    outputPath,
    *,
    fontSize=250,
    margin=20,
    features=None,
    variations=None,
    backendName=None,
):
    font = BlackRendererFont(fontPath)
    glyphNames = font.glyphNames

    scaleFactor = fontSize / font.unitsPerEm

    buf = hb.Buffer()
    buf.add_str(textString)
    buf.guess_segment_properties()

    if variations:
        font.setLocation(variations)

    hb.shape(font.hbFont, buf, features)

    infos = buf.glyph_infos
    positions = buf.glyph_positions
    glyphLine = buildGlyphLine(infos, positions, glyphNames)
    bounds = calcGlyphLineBounds(glyphLine, font)
    bounds = scaleRect(bounds, scaleFactor, scaleFactor)
    bounds = insetRect(bounds, -margin, -margin)
    bounds = intRect(bounds)
    if outputPath is None:
        suffix = ".svg"
    else:
        suffix = os.path.splitext(outputPath)[1].lower()
    if backendName is None:
        if suffix == ".svg":
            backendName = "svg"
        else:
            backendName = "skia"
    surfaceClass = getSurfaceClass(backendName, suffix)
    if surfaceClass is None:
        raise BackendUnavailableError(backendName)

    surface = surfaceClass()
    with surface.canvas(bounds) as canvas:
        canvas.scale(scaleFactor)
        for glyph in glyphLine:
            with canvas.savedState():
                canvas.translate(glyph.xOffset, glyph.yOffset)
                font.drawGlyph(glyph.name, canvas)
            canvas.translate(glyph.xAdvance, glyph.yAdvance)

    if outputPath is not None:
        surface.saveImage(outputPath)
    else:
        import tempfile

        with tempfile.NamedTemporaryFile(suffix=".svg") as tmp:
            surface.saveImage(tmp.name)
            with open(tmp.name, "rb") as f:
                svgData = f.read().decode("utf-8").rstrip()
        print(svgData)


def buildGlyphLine(infos, positions, glyphNames):
    glyphLine = []
    for info, pos in zip(infos, positions):
        g = GlyphInfo(
            name=glyphNames[info.codepoint],
            gid=info.codepoint,
            xAdvance=pos.x_advance,
            yAdvance=pos.y_advance,
            xOffset=pos.x_offset,
            yOffset=pos.y_offset,
        )
        glyphLine.append(g)
    return glyphLine


def calcGlyphLineBounds(glyphLine, font):
    bounds = None
    x, y = 0, 0
    for glyph in glyphLine:
        glyphBounds = font.getGlyphBounds(glyph.name)
        if glyphBounds is None:
            continue
        glyphBounds = offsetRect(glyphBounds, x + glyph.xOffset, y + glyph.yOffset)
        x += glyph.xAdvance
        y += glyph.yAdvance
        if bounds is None:
            bounds = glyphBounds
        else:
            bounds = unionRect(bounds, glyphBounds)
    return bounds


class GlyphInfo(NamedTuple):
    name: str
    gid: int
    xAdvance: float
    yAdvance: float
    xOffset: float
    yOffset: float
